
/*			Main Entry Point for the CFS-C Classifier System.

This file,	CFSC.C, contains the main entry point for the
CFS-C classifier system  (Copyright 1986,1988 Rick L. Riolo), Version XV 0.1 (-joke/J.E.) 

To run the system, it must be linked to some domain-specific subroutines to implement
an environment, e.g., the FSW1 environment.  If the runable file is called fsw1, then use

	fsw1  [ -w ] [ -b batchfile ] [ -s stringsize ]  [ load-file ]
	
Where	'-w' makes the system start the Xv-window immediately without 
	    prompting the user for an input from stdin. Releasing the
	    window terminates the application, then.
	'-b batchfile' also never prompts the user but reads any 
	    commands from <batchfile> and - after each termination
	    of the Classify-function - prints the number of positive 
	    rewards to the file '<batchfile>.out'. The application
	    terminates when all commands are read and executed.
	    (See also the shellscript 'batch' which uses this option).
	'-s stringsize' sets the Variable STRNGSZ to the given value.
	    In earlier Versions, STRNGSZ used to be a constant, so that
	    binding CFSC to a different environment or even the same en-
	    vironment with different detector-/effector sizes took to
	    recompile all CFSC objects resp. the whole CFSC library.
	    This is no longer necessary. NOTE that there is still a 
	    maximum stringsize compiled into the executables, called
	    STRNGMAXSZ (set to 64). If the problem size exceeds this maximum,
	    though, a recompilation is needed. 
	    -s defaults to 16 if not mentioned.

<load-file> is also optional; it specifies the full path/filename for a file that was
created with the CFS-C SAVESYS command.  It basically contains the state of the system
at the time the command was issued, and when the system is started with such a file it
starts in that state.

This file contains these subroutines:

	main		the main entry for the system.
	DoCmd		execute a user command.
	ExecCmds	read and execute commands from a file.
	Classify	the major-cycle loop.
	StartStp,EndStep	initialize and clean-up before/after a major-cycle step.
	Init_CFS,InitVT 	initialize the system.
	InitCfLs	initialize the classifier list.

	LoadSys 	load entire system from a file.
	ReadVars	read in core system variables
	LinkCMs 	link classifiers and messages after a "re-start" LoadSys().
	GetMsgAd	get address for message with specified Msg_Id.
	GetCfAd 	get address for classifier with specified Cf_Id.

	SaveSys 	write entire system state to file for use by LoadSys().
	WrtCons 	write core system constants
	ReadEffs	read effector conditions, names, etc, and store in effector nodes (EffNode's).
	WrtEffs 	write EffNode contents in format ready to be read by ReadEffs().

	InitDate	initialize the RunDate, RunTime and SaveSyFN variables.
**/

char CRight[] = "Copyright 1986,1988 Rick L. Riolo";
char ShortCRight[] = "(C)'86,'88 R.L.Riolo";
#include <stdio.h>

#ifdef CFSC_XVIEW
#include <sys/param.h>
#include <sys/types.h>
#include <xview/xview.h>
#include <xview/canvas.h>
#include <xview/panel.h>
#include <xview/scrollbar.h>
#include <xview/svrimage.h>
#include <xview/termsw.h>
#include <xview/text.h>
#include <xview/tty.h>
#include <xview/xv_xrect.h>
#endif

#include	"config.h"		/* set a switch indicating compiler/machine being used */

#include	"utility.h"		/* General purpose constants */
#include	"utility.ext"		/* Function external definitions */
#include	"core.ext"		/* extern's for major data structures */
#include	"usercmd.ext"		/* some stuff for processing user commands */
#include	"cfsio.ext"		/* some I/O declarations */
#include	"cfmsgio.ext"		
#include	"dsclearn.ext"		/* Used in StartStp() to reset vars */
#include	"grandcfs.ext"		/* For extern decl of GnRndCfs() */
#include	"cfsutil.ext"
#include	"bblearn.ext"
#include	"display.ext"
#include	"cfsc.def"

#ifdef CFSC_XVIEW
#include	"cfsc_ui.h"
#endif

char Version[] = "XV 0.1 (-joke/J.E.)";	/* Version number	****  How to include the compilation date?? -joke says: use __DATE__ ;-) **** */
char RunDate[] = "00/00/00";		/* Date of run. */
char RunTime[] = "00:00";		/* Time of run. */
int  CFSC_BATCH;

/*******************************

Main		The main entry point for the system.

This calls functions to initialize the system, and then enters a loop to process user commands.
Each command is read and then processeed by appropriate function.

*******/

FILE *myf;

#if !MACLSC
VOID 
main (argc, argv)
#else
VOID 
_main (argc, argv)
#endif
    int argc;
    char *argv[];
{
    int CmdCode,i,win;
    char CmdPars[CMDPARSZ + 1];
    char CFSC_BATCH_FILENAME[FNAMESZ];
    char CFSC_BATCH_OUTFILENAME[FNAMESZ];

#if  MACAZTEC
    InitFonts ();
    TextFont (4);			/* Monoco 9 point seems pretty good */
    TextSize (9);
#endif

#ifdef CFSC_XVIEW
    xv_init(XV_INIT_ARGC_PTR_ARGV, &argc, argv, NULL);
/*    notify_do_dispatch(); */
#endif
    
    CFSC_BATCH=FALSE;
    STRNGSZ=16;
    win=FALSE;
    for (i=1;i<argc;i++) {
	if (strncmp(argv[i],"-b",2)==0) {
	    strcpy(CFSC_BATCH_FILENAME,argv[i]+2);
	    CFSC_BATCH=TRUE;
	}
	if (strncmp(argv[i],"-w",2)==0) {
	    win=TRUE;
	}
	if (strncmp(argv[i],"-s",2)==0) {
	    STRNGSZ=atoi(argv[i]+2);
	    MutdSpec=(1.0 / (2.0 * STRNGSZ)); /* This cannot be done
			    by an initializer. */
	    if (STRNGSZ==0 || STRNGSZ>STRNGMAXSZ) {
		sprintf (GOutBuff, "\nCFS-C: stringsize %d is too large or 0 (%d is maximum).\n If too large, change STRNGMAXSZ and recompile CFS-C.\n",STRNGSZ,STRNGMAXSZ);
		WriteStd (GOutBuff);
		exit;
	    }
	}
    }
    if (*argv[argc-1]!='-' && argc!=1)
	Init_CFS (argv[argc-1]);
    else Init_CFS ("");

    if (win) {
#ifdef CFSC_XVIEW
	editCfsc(NULL);
	xv_main_loop(editCfscSubframe);
#endif
    } else {
	if (!CFSC_BATCH) {
	    while (!StopFlag) {
		CmdCode = GUserCmd (CmdPars);	/* Get code for command. */
		DoCmd (CmdCode, CmdPars);	/* Execute the command (may set StopFlag TRUE) */
	    }
	}
	if (CFSC_BATCH) {
	    strcpy(CFSC_BATCH_OUTFILENAME,CFSC_BATCH_FILENAME);
	    strcat(CFSC_BATCH_OUTFILENAME,".out");
	    myf=fopen(CFSC_BATCH_OUTFILENAME,"w");
	    ExecCmds(CFSC_BATCH_FILENAME);
	    fclose(myf);
	}
    }

    if (LogFlg) {
	LogFlg = OFF;
	CloseFILE (LogFName, LogFName, LogFILE);
	sprintf (GOutBuff, "\nLogging on '%s' terminated.\n", LogFName);
	WriteStd (GOutBuff);
    }
    if (GenFlg) {
	GenFlg = OFF;
	CloseFILE (GenFName, GenFName, GenFILE);
	sprintf (GOutBuff, "\nGeneology recording on '%s' terminated.\n", GenFName);
	WriteStd (GOutBuff);
    }
}					/* main */

/*******************************

DoCmd		  Execute Command entered by user.

  CmdCode	  Code for command, returned by GUserCmd().

  CmdPars	  Parameters included on command.

  Note that this may set the global variable StopFlag TRUE to stop the run.

******/

VOID 
DoCmd (CmdCode, CmdPars)
    unsigned int CmdCode;
    char *CmdPars;
{
    extern VOID DoEnvCmd();
    extern int LoadEnv();   /*	This must be declared external in this place,
				because the environment which CFSC is bound
				to is yet unknown. */

    switch (CmdCode) {
     case LSYS_CC:
	 DoLoadSys(CmdPars);
	 break;
#ifdef CFSC_XVIEW
     case WIN_CC:
	 editCfsc(NULL);
	 xv_main_loop(editCfscSubframe);
	 break;
#endif
     case STOP_CC:
	 StopFlag = TRUE;
	 break;
     case LOADC_CC:
	 LoadCfs (CmdPars, 'r');	/*  Load classifiers (replace old) */
	 break;
     case LOADM_CC:
	 LoadMsgs (CmdPars, 'r', 'c');	/* Load messages (replace old) into current list. */
	 break;
     case LOADE_CC:
	 LoadEnv (CmdPars);		/*  user supplies a file or use default */
	 break;
     case GENRC_CC:
	 GnRndCfs (CmdPars);		/*  generate random classifiers: par="how_many,meanStr,meanBR" */
	 break;
     case DISPL_CC:
	 DoDisplay (CmdPars);		/*  [parameter [,parameter]+]	 */
	 break;
     case SET_CC:
	 Set (CmdPars);			/*  parameter	value  */
	 break;
     case CLASS_CC:
	 Classify (CmdPars);		/* Run model major cycle. CmdPars indicates how may steps. */
	 if (CFSC_BATCH) 
	    fprintf(myf,"%d\n",TOTSyRwP);    /* Export No of rewarded steps as a performance measure */
	 break;
     case STEP_CC:
	 Classify ("1");		/* special case of RUN with  CmdPars = '1'  */
	 break;
     case CMDS_CC:
	 ExecCmds (CmdPars);		/* read and execute commands from file */
	 break;
     case ECMD_CC:
	 DoEnvCmd (CmdPars);		/* Do an environment-specific command */
	 break;
     case SAVES_CC:
	 SaveSys (CmdPars);		/* Save entire system to file */
	 break;
     case MDCF_CC:
	 Mod_Cf (CmdPars);		/* prompts for cf-id and new value */
	 break;
     case MDMSG_CC:
	 Mod_Msg (CmdPars);		/* Prompts for msg-id and new value */
	 break;
     case APCF_CC:
	 LoadCfs (CmdPars, 'a');	/* Same as Load, except append classifers to current list. */
	 break;
     case APMSG_CC:
	 LoadMsgs (CmdPars, 'a', 'c');	/* Same as Load, except append messages to current list.  */
	 break;
     case HELP_CC:
	 Help (CmdPars);
	 break;
     case NULL_CC:
	 break;				/* Don't do anything for a null line */
     default:
	 WriteStd ("\n\nIllegal command. For help, enter HELP .\n");
	 break;
    }

}					/*  DoCmd	*/


/*******************************

ExecCmds	  Execute commands from a file until EOF reached.

  FDName	  Null-terminated string name of file from which commands are to be read.

******/

VOID 
ExecCmds (FDName)
    char *FDName;
{
    int retlen, CmdCode, cnt;
    char CmdPars[CMDPARSZ + 1], *lptr, Command[CMDNAMSZ + 1];	/* room for command name and \0 */
    FILE *fptr;

    if ((fptr = GetFILE (FDName, "", READMODE)) == NULL)
	WriteStd (" No commands to execute.\n");
    else {
	while (ReadS (GInBuff, GIBUFFSZ, fptr, &retlen) != EOF) {
	    if (EchoFlg) {
		WriteStd (GInBuff);
		WriteStd ("\n");
	    }
	    for (lptr = GInBuff; *lptr == ' '; ++lptr);	/* skip leading blanks */

	    for (cnt = 0; *lptr != COMCHAR && *lptr != ' ' && *lptr != '\0' && cnt <= CMDNAMSZ; ++lptr, ++cnt)
		Command[cnt] = isupper (*lptr) ? tolower (*lptr) : *lptr;
	    Command[cnt] = '\0';

	   /* Get CmdCode, or null line command code or error command code. */

	    if (cnt == 0)
		CmdCode = NULL_CC;
	    else if ((CmdCode = GetCmdCd (Command)) == ERROR)
		CmdCode = ERROR_CC;
	    else {
		while (*lptr == ' ')
		    ++lptr;		/* skip blanks before parameter */

		for (cnt = 0; *lptr != COMCHAR && *lptr != '\0' && cnt <= CMDPARSZ; ++cnt, ++lptr)
		    CmdPars[cnt] = isupper (*lptr) ? tolower (*lptr) : *lptr;
		CmdPars[cnt] = '\0';
	    }

	    DoCmd (CmdCode, CmdPars);

	}				/* more lines in file */

	CloseFILE (FDName, "", fptr);

    }					/* file was opened */

}					/* ExecCmds */


/*******************************

Classify	The major cycle.

On entry, StepsPar tells how many cycle steps should be executed.
The while-statement keeps executing all the major cycle subroutines until
the specified number of steps have been done or until something  sets StopFlag TRUE.

At the end of each major-cycle, the CurMsgs becomes the OldMsgs list, the
NewMsgs (generated during that step) becomes the CurMsgs, and the space used by
the previous OldMsgs list is setup to be used for NewMsgs on the next step.

******/

int Classify (StepsPar)
    char StepsPar[];
{
    int err;
    unsigned int StopStep, Steps, savesd;

    GetUInt (StepsPar, &Steps, 0, ";", &err);
    if (err || Steps < 1) {
	sprintf (GOutBuff, "\nERR--illegal count for CLASSIFY: %s", StepsPar);
	WriteStd (GOutBuff);
	return (ERROR);
    }
    for (StopStep = CycleStp + Steps, StopFlag = FALSE; CycleStp < StopStep && !StopFlag; ++CycleStp) {

	StartStp ();			/* Initialize various variables for bidding, etc. */

	RdDetect ();			/* Get messages from detectors */

	GenCands ();			/* Generate lists of candidate classifiers/matches; Accumulate support */

	GenNMsgs ();			/* Generate bids, pick winning classifiers, generate new messages. */

	GenBehav ();			/* Mark those effectors satisfied by the new messages, and link the
									satifying messages to those effectors. Resolve conflicts, if any.
									Update candidate classifiers and new messages to reflect any messages
									deleted by the conflict resolution mechanism. Generate behavior from
									activated effectors, and get reward from enviornment.
								*/
	if (!ClampFlg) {
	    UpdCfStr ();		/* Update Classifier strengths using payoffs from Environment,	and taxes. */

	    Discover (); 		/* Check "discovery" trigger conditions, and apply algorithms if appropriate. */
	}
	EndStep ();			/* Exchange new, current, and old message lists, and do other clean-up jobs. */

	if (AutoDspl > 0)
	    AuDisplay ();		/* AutoDisplay whatever the user has requested. */

	if (CycleStp % SaveInt == 0) {	/* Save the state of the system */
	    if (!ClampFlg) {
		++CycleStp;		/* KLUDGE: Make CycleStp equal to what it will be at the start */
		SaveSys ("");		/* of the next cycle, when a user could enter a SAVESYS command. */
		--CycleStp;		/* undo the kludge! */
	    }
	}
    }					/* major cycle */

    if (ClampFlg)			/* if we are clamped, de-increment CycleStp */
	--CycleStp;			/* ***NOTE WELL*** You must STEP when ClampFlg is on to keep CycleStp right */

    return (OK);

}					/* Classify */

/*******************************

StartStp	  Initialize variables for bidding, etc., at start of major-cycle step.

******/

int StartStp ()
{
    struct CfNode *cp;
    struct MsgNode *mp;
    register int mcnt;

    TotCCSup = TotBBid = TotCCBid = HiCCBid = 0;
    LowCCBid = BIGFLOAT;
    NmCandCf = NmMtchs = NmCfWon = NmCfPost = 0;

    ActCfPNm = 0;			/* Get ready to store active classifiers in prev active list */

    for (cp = CandCfs; cp != NULL; cp = cp -> NxtCand) {
	if (cp -> NmPost > 0) {		/* Its active, so store in prev list */
	    cp -> NmPost = cp -> PstEfMsg = 0;
	    ActCfPrv[ActCfPNm].ActCf = cp;	/* store pointer to it, and */
	    ActCfPrv[ActCfPNm].ChngStr = cp -> ChngStr;	/* change in strength this step */
	    ++ActCfPNm;			/* and increment count */
	}
	cp -> NmMtch = cp -> NmProd = 0;
    }

    CandCfs = NULL;

    for (mp = CurMsgs, mcnt = 0; mcnt <= LastCMsg; ++mp, ++mcnt) {
	mp -> MsgUsed = FALSE;
	mp -> MsgMtchd = FALSE;
    }

   /*	Make current active cf list into previous, and init current
			(This is used by DscACPC() in DSCOPS2.C)
		*/

    return (OK);

}					/* StartStp */


/*******************************

EndStep 	  Do clean-up jobs at the end of a major-cycle step.

	Exchange new, current, and old message lists.

	Update some global state variables (eg, AveCfStr - average strength of all classifiers).

******/

int EndStep ()
{
    struct MsgNode *oldlist;

    TOTNmBid += NmCandCf;		/* Totals for classifiers	*/
    TOTNmWin += NmCfWon;
    TOTCfPst += NmCfPost;
    TOTMtch += NmMtchs;			/* Total for matches, i.e., potential matches */
    TOTMsPst += NmNMsgs;		/* Total of messages produced (after effector conflict resolution) */

    oldlist = OldMsgs;			/* save pointer to those nodes */

    OldMsgs = CurMsgs;			/* Make current list the old list. */
    LastOMsg = LastCMsg;
    NmOMsgs = NmCMsgs;
    NmOMsgIn = NmCMsgIn;

    CurMsgs = NewMsgs;			/* Make new messages current */
    LastCMsg = LastNMsg;
    NmCMsgs = NmNMsgs;
    NmCMsgIn = NmNMsgIn;
    NextCMsg = NextNMsg;

    NewMsgs = oldlist;			/* Use (saved) old-message space for new list */
    NextNMsg = NewMsgs;
    NmNMsgs = NmNMsgIn = 0;
    LastNMsg = -1;			/* NewMsgs list is empty to start a cycle-step. */

    if (NmCfs > 0)

#if SUN3				/* kludge for SUN compiler */
	AveCfStr = TotCfStr / (int) NmCfs;
#else
	AveCfStr = TotCfStr / NmCfs;
#endif

    else
	AveCfStr = 0;

    return (OK);

}					/*  EndStep  */


#if  MPWC
#define  __SEG__ CFSCINIT
#endif

/*******************************

Init_CFS	  Initialize the classifier system.

Then get memory for the message lists, classifiers, and effectors,
and do some preliminary initialization of variables related to those data structures.

Then if RunFName is not a null string (""), initialize EVERYTHING from that file (if its openable).
Otherwise if RunFName  is a null string,
a. If STARTUP.CFS is openable file, initialize EVERYTHING from it.
b. Otherwise, just use defaults for variables and then load messages, classifiers,
	and the environment from the default files for each (see CFSIO.DEF).

******/

int Init_CFS (RunFName)
    char *RunFName;
{
    int i, retlen, err, err1;
    char *fname, *cp;
    FILE *fptr;
    struct MsgNode *cmp, *nmp, *omp;
    extern int LoadEnv();

#if  M_I86HM
    void *halloc ();			/* Huge model for MS C */
    unsigned int cfnodesz, cfnodeaz;
    long unsigned int cflstsz;
#endif

#if ( CI86 || CBELLMTS )
    extern FILE *StdOFILE, *DisOFILE;
    StdOFILE = DisOFILE = stdout;
#endif

    InitDate ();			/* Get RunDate and RunTime, setup SaveSyFN */

   /* Set up message list size limits:
    NmDMsgMx - setable number of spots on message list reserved for detector messages.
    NmIMsgMx - maximum number of 'internal' (non-detector) messages, which is all
    not reserved for detector messages (user/program setable).
    Thus ( NmDMsgMx + NmIMsgMx ) better be <= MSGLSTSZ  */

    if (NmIMsgMx + NmDMsgMx != MSGLSTSZ) {
	WriteStd ("\n\nNmIMsgMx and NmDMsgMx values in CORE.DEF not equal to MSGLSTSZ.");
	WriteStd ("\nSetting NmIMsgMx to 24 and NmDMsgMx to whatever is left!\n");
	NmIMsgMx = min (24, MSGLSTSZ);
	NmDMsgMx = MSGLSTSZ - NmIMsgMx;
    }
   /* Get space for message lists and initialize them */

    CurMsgs = (struct MsgNode *) calloc (MSGLSTSZ, sizeof (struct MsgNode));
    if (CurMsgs == NULL) {
	WriteStd ("\nERR in Init_CFS while getting memory for CurMsgs - calloc returned NULL!");
	exit (1);
    }
    NewMsgs = (struct MsgNode *) calloc (MSGLSTSZ, sizeof (struct MsgNode));
    if (NewMsgs == NULL) {
	WriteStd ("\nERR in Init_CFS while getting memory for NewMsgs - calloc returned NULL!");
	exit (1);
    }
    OldMsgs = (struct MsgNode *) calloc (MSGLSTSZ, sizeof (struct MsgNode));
    if (OldMsgs == NULL) {
	WriteStd ("\nERR in Init_CFS while getting memory for OldMsgs - calloc returned NULL!");
	exit (1);
    }
    for (i = 1, cmp = CurMsgs, nmp = NewMsgs, omp = OldMsgs; i <= MSGLSTSZ; ++i, ++cmp, ++nmp, ++omp) {
	cmp -> Msg_Id = nmp -> Msg_Id = omp -> Msg_Id = 0;
    }

    NextCMsg = CurMsgs;			/* Next messages will go at beginning. */
    NextOMsg = OldMsgs;
    NextNMsg = NewMsgs;

   /* Get space for classifiers and initialize them */

#if  M_I86HM
    cflstsz = CFLSTSZ;
    cfnodesz = sizeof (struct CfNode);
    cfnodeaz = cfnodesz + 6;		/* **NOTE**  Must fix to be 2**i ***  */
    CfLst = (struct CfNode *) halloc (cflstsz, cfnodeaz);
    if (CfLst == NULL) {
	WriteStd ("\nERR: Init_CFS, for CfLst: halloc return NULL!");
	sprintf (GOutBuff, "\n	 CFLSTSZ %d, CfNode alloc size %d (size %d)",
		 cflstsz, cfnodeaz, cfnodesz);
	WriteStd (GOutBuff);
	exit (1);
    } else {
	sprintf (GOutBuff, "\n  [** Using halloc: CFLSTSZ %ld, CfNode alloc size %d (real size %d) **]",
		 cflstsz, cfnodeaz, cfnodesz);
	WriteStd (GOutBuff);
    }
#else
    CfLst = (struct CfNode *) calloc (CFLSTSZ, sizeof (struct CfNode));
    if (CfLst == NULL) {
	WriteStd ("\nERR: Init_CFS, for CfLst: calloc return NULL!");
	exit (1);
    }
#endif

    InitCfLs ();

   /* Set up array of unsigned int's, so that i-th bit of i-th entry is set to 1, other bits 0. */

    for (i = 0; i < INTSZ; ++i) {
	ORMBits[i] = 1 << i;
    }
	

    sprintf (GOutBuff, "\nCFC-C Version %s. %s. RunDate %s (%s).", Version, CRight, RunDate, RunTime);
    WriteStd (GOutBuff);

#if  M_I86HM
    cflstsz = CFLSTSZ;
    cflstsz *= cfnodeaz;
    sprintf (GOutBuff, "\n[String Sz %d, MsgNode Sz %d, CfNode Sz %d, CFLSTSZ %d, ]",
	     STRNGSZ,sizeof (struct MsgNode), sizeof (struct CfNode), CFLSTSZ);
    WriteStd (GOutBuff);
    sprintf (GOutBuff, "CfLst size %lu].", cflstsz);
#else

#if ( LATTICEC && !__HIGHC__ )
    sprintf (GOutBuff, "\n[String Sz %d, MsgNode Sz %d, CfNode Sz %d, CFLSTSZ %d, CfLst sz %u, start %x, memavl %d].",
	     STRNGSZ,sizeof (struct MsgNode), sizeof (struct CfNode), CFLSTSZ, CFLSTSZ * sizeof (struct CfNode), CfLst, (unsigned)_memavl ());
#else
    sprintf (GOutBuff, "\n[String Sz %d, MsgNode Sz %d, CfNode Sz %d, CFLSTSZ %d, CfLst size %u, start %x].",
	     STRNGSZ,sizeof (struct MsgNode), sizeof (struct CfNode), CFLSTSZ, CFLSTSZ * sizeof (struct CfNode), (unsigned)CfLst);
#endif

#endif

    WriteStd (GOutBuff);

   /* If there is a RunFName, try to open it. Otherwise try to open the default startup file. */

    if (RunFName[0] != '\0') {
	if ((fptr = GetFILE (RunFName, RunFName, READMODE)) == NULL)
	    exit (1);
	else
	    fname = RunFName;		/* RunFName is open, so will load from it */
    } else if ((fptr = fopen (LoadSyFN, READMODE)) != NULL)
	fname = LoadSyFN;

    if (fptr != NULL) {			/* one of the startup files is open, so... */
	sprintf(GOutBuff,"\nInitializing from startup file %s...\n",fname);
	WriteStd (GOutBuff);
	LoadSys (fname, fptr);		/* use it to initialize the system */
    }
    else {				/* neither open, so load from separate (default) files */
	LoadCfs (CfInFN, 'r');
	LoadMsgs (MsgInFN, 'r', 'c');
	LoadEnv (EnvInFN); 
    }

    InitRand ();			/* init random number generator (if needed) */

    return (OK);

}					/* Init_CFS  */


/*******************************

InitCfLs - initialize classifier list.

	On entry, CfLst points to block of memory big enough for CFLSTSZ nodes.
	
******/

VOID 
InitCfLs ()
{
    register int i, j;

    for (i = 1, CurCfs = CfLst; i <= CFLSTMX; ++i, ++CurCfs) {
	CurCfs -> Cf_Id = 0;
	for (j = 0; j < INTPRSTRMAX; ++j) {
	    CurCfs -> Cnd1Bits[j] = 0;
	    CurCfs -> Cnd1DCs[j] = 0;
	    CurCfs -> Cnd2Bits[j] = 0;
	    CurCfs -> Cnd2DCs[j] = 0;
	    CurCfs -> ActBits[j] = 0;
	    CurCfs -> ActDCs[j] = 0;
	}
	for (j = 0; j < INTPRML; ++j)
	    CurCfs -> MMBitsC1[j] = CurCfs -> MMBitsC2[j] = 0;
	CurCfs -> NxtCand = NULL;
	CurCfs -> NmMtch = CurCfs -> NmMMC1 = CurCfs -> NmMMC2 = CurCfs -> NmProd = CurCfs -> NmPost = 0;
	CurCfs -> TotNmBid = CurCfs -> TotMtch = CurCfs -> TotProd = CurCfs -> TotPost = 0;
	CurCfs -> TotEMtch = CurCfs -> TotEAct = CurCfs -> TotPosRw = CurCfs -> TotNegRw = 0;
	CurCfs -> ChngStr = 0;
	CurCfs -> ReplFlg = CurCfs -> NoBBFlg = CurCfs -> NoRplFlg = CurCfs -> NoPrnFlg = FALSE;
    }

    NmCfs = 0;				/* No classifiers yet. */
    NxtFrCf = CfLst;			/* Next free node starts at beginning of memory-block. */
    CurCfs = CfLst;			/* When classifiers added, they will start here. */
    StrSrtCf = NULL;			/* Non sorted on strength yet. */

    TotCfStr = AveCfStr = TotCfBR = AveCfBR = 0.0;

}					/* InitCfLs */


/*************
DoLoadSys
*************/
void DoLoadSys(char *Par) {
    char *c;
    FILE *f;
    while (*Par==' ') Par++;
    c=Par;
    while (*c!=' ') c++;
    *c='\0';
    if ((f=fopen(Par,"r"))==NULL) {
	WriteStd ("Couldn't find file\n");
	return;
    }
    LoadSys(Par,f);
}

/*******************************

LoadSys 	  Load system start-state from the specified (open) file.

  Just call subroutines to load the various system parts, i.e.,
  the messages lists, classifiers, environment state, and system variables.

******/

VOID 
LoadSys (FileName, FilePtr)
    char *FileName;
    FILE *FilePtr;
{
    extern VOID ReadEnv();

    ReadMsgs (FilePtr, 'r', 'o');	/* load the old message list */
    ReadMsgs (FilePtr, 'r', 'c');	/* load the current message list */
    ReadCfs (FilePtr, 'r');		/* load the  classifiers */
    LinkCMs ();				/* Replace Msg and Classifer id's with pointers (where appropriate) */

    ReadEnv (FilePtr);

    ReadVars (FilePtr);

    if (CloseFILE (FileName, "", FilePtr) == ERROR) {
	WriteStd ("--system not properly loaded.\n\n");
	exit (1);
    }
}					/* LoadSys */

/*******************************

ReadVars	Read values for setable core variables from a file.

FilePtr 	FILE pointer to already opened file/device.

Each input lines should be of the form:
	varname=value
where 'varname' is the name of a variable in the CVTable of core variables (see USERCMD.DEF),
and 'value' is the value to be assigned.
(Lines beginning with a semicolon are ignored as 'comment' lines.)

******/

VOID 
ReadVars (FilePtr)
    FILE *FilePtr;
{
    int retlen;

   /* Read lines from input, call SetVar to do the assignment. Stop on end-of-file or "END-VARIABLES" string. */

    while (ReadS (GInBuff, GIBUFFSZ, FilePtr, &retlen) != EOF) {
	if (*GInBuff == COMCHAR)	/* ignore comments */
	    continue;

	else if (strcmp (GInBuff, "END-VARIABLES") == 0)	/* thats all, folks  */
	    break;

	else
	    SetVar (GInBuff, 'l');	/* The 'l' means respect the CVLoad flag for the variable. */

    }

}					/* ReadVars  */


/*******************************

LinkCMs 	  Link messages and classifiers.

  When the system is saved by SaveSys(), the id's of messages and classifiers
  are written to indicate:
	- which classifier produced a message  (*Producer in MsgNodes).
	- which messages (if any) were used by that classifier to produce the message (*MtchMsg1 and 2 in MsgNodes).
  When the system is later loaded by LoadSys(), those id's are stored in the indicated pointer variables.

  LinkCMs() goes back and gets addresses for the messages and classifiers
  with those id numbers, and stores those addresses in the proper pointer variables.
  NOTE that the MtchMsg1's and 2's are messages in the OldMsgs list.

  This is all required so that the bucket-brigade will work after a SaveSys/LoadSys cycle.

******/

VOID 
LinkCMs ()
{
    unsigned int i;
    struct MsgNode *curmsg;

    for (i = 1, curmsg = CurMsgs; i <= NmCMsgs; ++i, ++curmsg) {
	if (curmsg -> Producer != NULL)
	    if ((curmsg -> Producer = GetCfAd ((unsigned int) curmsg -> Producer)) == NULL) {
		sprintf (GOutBuff, "LinkCMs couldn't find Producer for Msg %u. Storing NULL.\n",
			 curmsg -> Msg_Id);
		WriteStd (GOutBuff);
	    }
	if (curmsg -> MtchMsg1 != NULL)
	    if ((curmsg -> MtchMsg1 = GetMsgAd ((unsigned int) curmsg -> MtchMsg1, 'o')) == NULL) {
		sprintf (GOutBuff, "LinkCMs couldn't find MtchMsg1 for Msg %u. Storing NULL.\n",
			 curmsg -> Msg_Id);
		WriteStd (GOutBuff);
	    }
	if (curmsg -> MtchMsg2 != NULL)
	    if ((curmsg -> MtchMsg2 = GetMsgAd ((unsigned int) curmsg -> MtchMsg2, 'o')) == NULL) {
		sprintf (GOutBuff, "LinkCMs couldn't find MtchMsg2 for Msg %u. Storing NULL.\n",
			 curmsg -> Msg_Id);
		WriteStd (GOutBuff);
	    }
    }

}					/*  LinkCMs  */


/*******************************

GetMsgAd	  Get address of MsgNode that contains message with specified Id.

  Id		  Id of message to look for.

  List		  List to search. Currently this is always 'o' -- search the OldMsgs list.

  Return:	  Pointer to MsgNode found, or NULL (and print message) if none found.

******/

struct MsgNode *
GetMsgAd (unsigned int Id, char List)
{
    unsigned int i;
    struct MsgNode *mptr;

    if (List == 'o') {
	for (i = 1, mptr = OldMsgs; i <= NmOMsgs; ++i, ++mptr)
	    if (mptr -> Msg_Id == Id)
		return (mptr);
    } else
	WriteStd ("\n\nBUG (GetMsgAd): illegal List specified.\n\n");

    sprintf (GOutBuff, "\n\nERR (GetMsgAd): can't find message with Msg_Id %u.\n", Id);
    WriteStd (GOutBuff);
    return (NULL);

}					/* GetMsgAd */


/*******************************

GetCfAd 	  Get address of CfNode that contains classifer with specified Id.

  Id		  Id of classifier to look for.

  Return:	  Pointer to CfNode found, or NULL (and print message) if none found.

******/

struct CfNode *
GetCfAd (Id)
    unsigned int Id;
{
    unsigned int i;
    struct CfNode *cptr;

    for (i = 1, cptr = CurCfs; i <= NmCfs; ++i, ++cptr)
	if (cptr -> Cf_Id == Id)
	    return (cptr);

    sprintf (GOutBuff, "\n\nERR (GetCfAd): can't find classifier with Cf_Id %u.\n", Id);
    WriteStd (GOutBuff);
    return (NULL);

}					/* GetCfAd */


/*******************************

SaveSys 	  Save entire system state, in format that LoadSys can use.

  FDName	  Name of file to be used for output.
			  If its a null string (""), write to next autosave file.

******/

int 
SaveSys (FDName)
    char *FDName;
{
    FILE *fileptr;
    char savefn[FNAMESZ+1];
    extern VOID WrtEnv();

    while (*FDName == ' ')
	++FDName;			/* Skip blanks */

    if (*FDName == '\0') {		/* No file specified, so construct one */
	sprintf (savefn, "%s%s.%03u", SaveSyPa, SaveSyFN, SaveFNum);	/* including the path and autosave number */
	SaveFNum++;			/* increment it now for next use */
	FDName = savefn;		/* and that's what we'll try to use. */
    }
    if ((fileptr = GetFILE (FDName, FDName, WRITEMODE)) == NULL)
	return (ERROR);

    fprintf (fileptr, ";  CONSTANTS\n");
    WrtCons (fileptr);

    fprintf (fileptr, ";  OLD-MSGLIST\n");
    WrtMsgs (fileptr, SAVEFMT, 'o');

    fprintf (fileptr, ";  CUR-MSGLIST\n");
    WrtMsgs (fileptr, SAVEFMT, 'c');

    fprintf (fileptr, ";  CLASSIFIERS\n");
    WrtCfs (fileptr, SAVEFMT);

    WrtEnv (fileptr, SAVEFMT);

    fprintf (fileptr, ";  VARIABLES\n");
    PrnVars (fileptr, SAVEFMT);   /* SAVEFMT is ignored so far: there is only one format */
    fprintf (fileptr, "END-VARIABLES\n");

    CloseFILE (FDName, FDName, fileptr);

    return (OK);

}					/* SaveSys */

/*******************************

WrtCons 	  Write system constants and other non-settable items that serve
			  to define the state of the classifier system during a run.

  FilePtr	  FILE pointer to open file/device.

******/

VOID 
WrtCons (FilePtr)
    FILE *FilePtr;
{

    fprintf (FilePtr, ";  Run-Date:  %s.  Run-Time:  %s.\n", RunDate, RunTime);

    fprintf (FilePtr, ";  STRNGSZ  (message-length):                    %d\n", STRNGSZ);
    fprintf (FilePtr, ";  STRNGMAXSZ  (Max. message-length):            %d\n", STRNGMAXSZ);
    fprintf (FilePtr, ";  MSGLSTSZ (max. number of msgs):               %d\n", MSGLSTSZ);
    fprintf (FilePtr, ";  CFLSTSZ  (max. number of classifers):         %d\n", CFLSTSZ);
    fprintf (FilePtr, ";  EFFLSTSZ (max. number of effectors):          %d\n", EFFLSTSZ);
    fprintf (FilePtr, ";Conditional-compilation switch settings:\n");
    fprintf (FilePtr, ";  CTEST   (test code):						 %d\n", CTEST);
    fprintf (FilePtr, ";  CDSC_ALL (Compile all discovery code):	   %d\n", CDSC_ALL);
    fprintf (FilePtr, ";  CRANDCFS (Compile GENRANDCFS command code):  %d\n", CRANDCFS);

    fprintf (FilePtr, ";END-CONSTANTS\n");

}					/*  WrtCons  */


/*******************************

ReadEffs	  Read effector condition-strings from a file, and initialize effectors.

  FilePtr	  FILE pointer to open file/device with effector strings, one per line,
			  followed by a line with the string "END-EFFS" starting in column one.
			  The lines should look like these:
				;  effector-condition,	effector-name,	Total-Matches, TotalActivations
				m0101 0101 0101 0100, Guess Letter, 0, 0
				~#### #### #### ####, Default Eff, 0, 0
				END-EFFS

  Returns	  The number of effectors read and loaded.

******/

int 
ReadEffs (FilePtr)
    FILE *FilePtr;
{
    char buff[STRNGMAXSZ + 1], name[EFFNAMSZ + 1], *lptr, *rptr;
    int retlen, nmefload, etotmtch, etotact, err, ecndtype;

    nmefload = 0;
    while (ReadS (GInBuff, GIBUFFSZ, FilePtr, &retlen) != EOF) {
	if (*GInBuff == COMCHAR)
	    continue;
	if (strcmp (GInBuff, "END-EFFS") == 0)
	    break;
	else if (nmefload >= EFFLSTSZ) {
	    sprintf (GOutBuff, "\n\nWARNING (ReadEff): you are trying to load more effector conditions (%d)", nmefload);
	    WriteStd (GOutBuff);
	    sprintf (GOutBuff, "\n than the maximum allowed (%d).", EFFLSTSZ);
	    WriteStd (GOutBuff);
	    continue;
	}
	for (lptr = rptr = GInBuff; *rptr != '\0'; ++rptr)	/* strip blanks */
	    if (*rptr != ' ')
		*lptr++ = *rptr;
	*lptr = '\0';

       /* check out this line, store parts in local variables */

	lptr = GInBuff;
	if (*lptr == '~') {
	    ecndtype = CNOTMATCH;	/* NOT-MATCH condition 2 */
	    ++lptr;			/* move past the ~ */
	} else {
	    ecndtype = CMATCH;		/* MATCH Type */
	    if (*lptr == 'm')
		++lptr;			/* move past the m if its there */
	}

	lptr = StpTok2 (lptr, buff, STRNGMAXSZ + 1, ",");
	if (!IsCondAct (buff)) {
	    sprintf (GOutBuff, "\nERR (ReadEff): illegal effector condition:\n'%s'\nnot loaded.\n", GInBuff);
	    WriteStd (GOutBuff);
	    continue;
	}
	lptr = StpTok2 (lptr, name, sizeof (name), ",");

	lptr = GetInt (lptr, &etotmtch, 0, ",", &err);
	if (err == TRUE) {
	    sprintf (GOutBuff, "\nWARNING (ReadEff): Illegal ETotMtch on line:\n%s\nUsing 0.", GInBuff);
	    WriteStd (GOutBuff);
	}
	lptr = GetInt (lptr, &etotact, 0, ",", &err);
	if (err == TRUE) {
	    sprintf (GOutBuff, "\nWARNING (ReadEff): Illegal ETotAct on line:\n%s\nUsing 0.", GInBuff);
	    WriteStd (GOutBuff);
	}
       /* We got it all, so store in the next effector node */

	strcpy (EffLst[nmefload].EffName, name);
	ACndtoB (buff, EffLst[nmefload].ECndBit, EffLst[nmefload].ECndDCs);
	EffLst[nmefload].ETotMtch = etotmtch;
	EffLst[nmefload].ETotAct = etotact;
	EffLst[nmefload].ECndType = ecndtype;
	EffLst[nmefload].Eff_Id = nmefload + 1;	/* Id's assigned in order */
	++nmefload;
    }

    return (nmefload);

}					/* ReadEffs */


/*******************************

WrtEffs 	Write out effectors in format ready to be read by ReadEffs().

FilePtr 	FILE pointer to open file/device.

******/

VOID 
WrtEffs (FilePtr)
    FILE *FilePtr;
{
    struct EffNode *eptr;
    int i;
    char buff[STRNGMAXSZ + 1], m;

    for (eptr = EffLst, i = 1; i <= NmEffs; ++i, ++eptr) {

	BCndtoA (eptr -> ECndBit, eptr -> ECndDCs, buff);
	buff[STRNGSZ] = '\0';

	if (eptr -> ECndType == CMATCH)
	    m = 'm';
	else
	    m = '~';

	fprintf (FilePtr, "%c%s, %s, %u, %u\n", m, buff, eptr -> EffName, eptr -> ETotMtch, eptr -> ETotAct);

    }

    fprintf (FilePtr, "END-EFFS\n");

}					/* WrtEffs */


/*******************************

InitDate	Initialize RunDate, RunTime and SaveSyFN, based on time and date from system.
		
******/

VOID 
InitDate ()
{
    int month, day, hour, minute;

    GetDate (RunDate);			/*	mm/dd/yy format */
    GetCFSCTime (RunTime, 0);		/*	hh:mm	 format */

    RunDate[8] = '\0';			/* Just in case... */
    RunTime[5] = '\0';

   /*	Put into SaveSyFN a string of the form	"Smddhhmm.", where
				S		is constant, to make it easy to find all S files.
				m		is the RunDate month, in HEX
				dd		is the RunDate day.
				hh		is the RunTime hour.
				mm		is the RunTime minute.
			Also put that into the LogFName, except start with L and end LOG,
			and into the GenFName, with the suffix GEN.
		*/

    sscanf (RunDate, "%2d/%2d/", &month, &day);
    sscanf (RunTime, "%2d:%2d", &hour, &minute);

    sprintf (SaveSyFN, "S%1x%02d%02d%02d", month, day, hour, minute);
    sprintf (LogFName, "L%1x%02d%02d%02d.LOG", month, day, hour, minute);
    sprintf (GenFName, "G%1x%02d%02d%02d.GEN", month, day, hour, minute);

}					/* InitDate */
